! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: mosaicmerge_mod
!
!
! Authors:  Brendan Perry - NASA/GSFC (brendan.h.perry@nasa.gov)
!            Comprehensive rewrite and modularization for SAS-21
!           Steven L. Snowden - NASA/GSFC (original algorithm)
!           Kip Kuntz - NASA/GSFC/JHU (original algorithm)
!
! Description: Extended Source Analysis Software
! SLS 7 May 2009
!
! Comments:
!
! $Id: mosaicmerge_mod.f90,v 1.4 2009/05/26 12:53:39 aibarra Exp $
!-

!  Merges various observations into one large mosaic
!
!  merge_comp_xmm mosaics count, modeled background count, soft proton
!  count, and exposure images from multiple pointings into one image.  
!  The output projection is the zenith equal-area azimuthal, which can
!  be approximated for display purposes by using tangential-projection
!  software
!
!  INPUT (new, SAS-21):
!    dirfile: ASCII list of directories to be merged
!    type: Type of file to mosaic (1=fovim, 2=expim, 3=bkgim, 4=protim, 5=swcxim, 6=cheese, 7=mslcheese)
!    coordsys: Coordinate system (1/equ, 2/gal, 3/ecl)
!    crval1: Central Longitude (crval(1)) in decimal degrees
!    crval2: Central Latitude (crval(2)) in decimal degrees
!    pixelsize: 

!    xmmdir: Directory containing the ESAS calibration files
!    dirfile: Filename of directory list
!    coord: Coordinate Control:
!       1 => ecliptic   lon/lat
!       2 => equatorial ra/dec
!       3 => galactic   L/B (or Lii/Bii???)
!    crvaln: Central Coordinates, longitude and latitude
!    scale: Pixel Size (arc minutes)
!    compon: Component Selection
!       1 => count image
!       2 => exposure
!       3 => QPB counts
!       4 => SP counts
!       5 => SWCX counts
!       6 => Cheese masks from cheese or cheese-bands
!       7 => Cheese masks from make_mask_merge
!    iel: Lower energy limit for the energy band in eV
!    ieh: Upper energy limit for the energy band in eV
!    maskcon: Mask control
!       0 =>  No Masking
!       1 =>  Point Source Masking - Use either output from cheese or 
!             the combined band output from cheese-bands
!       11 => Point Source Masking - Use the soft band output from 
!             cheese-bands
!       12 => Point Source Masking - Use the hard band output from 
!             cheese-bands
!       2 =>  Good Area Masking
!       3 =>  Point Source Masking - Use the full band output from 
!             make-mask-merge
!       31 => Point Source Masking - Use the soft band output from 
!             make-mask-merge
!       32 => Point Source Masking - Use the hard band output from 
!             make-mask-merge
!     alpha: Spectral Scale Factor - Assumed spectrum for scaling 
!       exposure images.  1.0 < alpha < 2.4
!
!  Example: mosaicmerge caldb=/CalDB dirfile=dir.dat coord=2 
!                 crvaln1=194.6 crvaln2=27.7 pixelsize=0.0666667  
!                 component=1 elow=400 ehigh=750 maskcontrol=0
!                 alpha=1.7
!
!  OUTPUT:
!    Mosaicked image
!

module mosaicmerge_mod

  use dal
  use param 
  use errorHandling
  use gofutils_mod  ! from goflib, formerly in marq_mod
  use gofcoords_mod ! from goflib, formerly in marq_mod
  use gofio_mod     ! from goflib, formerly in subsn (SS)
  use esas_mod      ! from goflib, formerly in subsn (SS)

  implicit none

  private

  character(len=*), parameter :: &
       dirfilepar   = "dirfile", &
       typepar      = "type", &
       cmospar      = "cheesemosaictype", &
       coordpar     = "coordsys", &
       crval1par    = "crval1", &
       crval2par    = "crval2", & 
       pixelpar     = "pixelsize", &
       elowpar      = "elow", &
       ehighpar     = "ehigh", &
       alphapar     = "alpha", &
       xdimpar      = "xdim", &
       ydimpar      = "ydim", &
       maskpar      = "withmask", &
       !maskfilepar  = "maskfile", &
       cheesepar    = "withcheese", &
       cmaskpar     = "cheesemasktype"

  public mosaicmerge

contains

  !---------------------------------------------------------------------
  subroutine mosaicmerge
    !---------------------------------------------------------------------

    type(DataSetT) :: in_S, inOOT_S, out_S, inMASK_S
    type(ArrayT)   :: in_A, inOOT_A, out_A, inMASK_A
    type(TableT)   :: in_T, inOOT_T, out_T, inMASK_T
    character(len=1024) :: in_F, inOOT_F, curdir, ctemp, dir_F, mask_F
    character(len=1024), allocatable, dimension(:) :: dirs
    character(len=256) :: basename, baseoot, out_F
    character(len=2) :: coordsys
    character(len=4) :: cmostype, cmasktype, ctyp
    character(len=5) :: ctypn, ctypo, ctypOO, celow, cehigh, cxdim, cydim
    character(len=8) :: filter, filtN, filtO, filtOO, units, inst, instN, instO, &
         expidN, expidO, instOO, expidOO
    character(len=8), allocatable, dimension(:) :: prefixes
    character(len=8), dimension(2) :: ctype, ctypeN, ctypeO, cunitN, cunitO,  cunitOO
    character(len=9) :: types(7) = [character(len=9) :: &
         "fov", "exp", "bkg", "prot", "swcx", "cheese", "mslcheese"]
    character(len=10) :: obsidN, obsidO, obsidOO, radecsysN, radecsysO, radecsysOO
    character(len=25) :: smodeN, smodeO, smodeOO
    character(len=30) :: comp, compo, compoo
    character(len=60) :: com, comment
    integer :: xGood=900, yGood=900, xSize, ySize, xdim, ydim, elow, ehigh, &
         type, typeA, typeF
    integer :: elowN, elowO, ehighN, ehighO, elowOO, ehighOO, frmtimeN, frmtimeO, frmtimeOO, &
         revlN, revlO, revlOO
    integer :: lun, nn, nRows, nRowsN, nRowsO, nRowsOO, nPrefixes, lastnn
    integer :: bitpix, blockn, coord, gcount, i, ii, j, jj, k, kk, n, &
         ios, ipixcrd(2), islow, ishigh, itot, luno, maskcon, imCounts, totBack, &
         totCounts, totExpo, totProt, totSwcx, totOO
    integer, allocatable, dimension(:,:) :: ini4l
    integer :: err = 0
    integer :: imed = 0
    integer :: ipnobj = 0
    integer :: ithick = 0
    integer :: ithin = 0
    integer :: istat = 0
    integer :: status = 0
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32, fovim_JP, fovimoot_JP, temp_JP
    !
    logical :: clob, clobber, extend, nomask, simple
    logical :: oot, die, warndie, withpart, withprot, withswcx, withmask, withbin, &
         withsmoo, withoot, withcomb, withchee, exists, cast, pnpn, pn, cmos
    real(kind=single) :: lat, lon, pixSize, areaN, areaO, areaOO, backN, backO, backOO, &
         expoN, expoO, expoOO
    real(kind=single) :: alpha, bscale, bzero, fscal, ootsca, scale
    real(kind=single), allocatable, dimension(:,:) :: mosaic, fovim, fovimoot, mask ! OLDNEW: merg input inputo mask
    real(kind=single), dimension(:,:), pointer :: aDR32, fovim_RP
    real(kind=double), dimension(:,:), pointer :: aDR64
    real(kind=double), dimension(2) :: cdelt, cdeltn, cdelto, cdeltOO, crotaN, crotaO, crotaOO, &
         crpix, crpixn, crpixo, crpixOO, crval, crvaln(2), crvalo(2), crvalOO(2), pixcrd
    real(kind=double) :: crotn, croto, cdec, cra, dbin, dbout, & 
         ddeccen, delang, djul, dlin, dlout, dracen, dpdec, dpra, &
         field, xpixin, ypixin, ra_pntN, ra_pntO, ra_pntOO, dec_pntN, dec_pntO, dec_pntOO, &
         pa_pntN, pa_pntO, pa_pntOO, xpixdelt, xrefpix, ypixdelt, yrefpix
    real(kind=double) :: drad = 3.141592654D0
    !
    imCounts  = 0.
    totCounts = 0.
    totBack   = 0.
    totExpo   = 0.
    totProt   = 0.
    totSwcx   = 0.
    drad      = drad/180.D0
    !
    ! SS: The coords are year 2000
    !
    djul = 2451544.5D0
    !
    ! Get the input parameters
    !
    dir_F     = stringParameter(dirfilepar)
    type      = intParameter(typepar)
    coordsys  = stringParameter(coordpar)
    ! BHP: both number or letters are acceptable for coordsys
    if ((coordsys .eq. '1') .or. (coordsys(1:2) .eq. 'eq'))then
       coordsys = 'eq' ! equatorial
    elseif ((coordsys .eq. '2') .or. (coordsys(1:2) .eq. 'ga'))then
       coordsys = 'ga' ! galactic
    elseif ((coordsys .eq. '3') .or. (coordsys(1:2) .eq. 'ec'))then
       coordsys = 'ec' ! ecliptic
    else
       call error("badCoordsys","Coordsys param must be (1 or ec), (2 or eq), (3 or ga)")
    endif
    !
    crvalN(1) = realParameter(crval1par)
    crvalN(2) = realParameter(crval2par)
    pixSize   = realParameter(pixelpar)
    elow      = intParameter(elowpar)
    ehigh     = intParameter(ehighpar)
    alpha     = realParameter(alphapar)
    xdim      = intParameter(xdimpar)
    ydim      = intParameter(ydimpar)
    withmask  = booleanParameter(maskpar)
    withchee  = booleanParameter(cheesepar)
    !
    if (withmask .and. withchee) then
       call error("NoMaskANDCheese","Mosaicmerge can apply an ordinary mask OR cheese mask but not both.")
    endif
    if (withchee) then
       cmasktype = stringParameter(cmaskpar)
       cmasktype = to_lower(cmasktype)
       write(errstr,*)'Apply cheese masks?         ',withchee
       call message(AppMsg,Always)
       write(errstr,*)'Cheese mask type:           ',cmasktype
       call message(AppMsg,Always)
       if ((trim(adjustl(cmasktype))  .eq. "s") .OR. &
            (trim(adjustl(cmasktype)) .eq. "h") .OR. &
            (trim(adjustl(cmasktype)) .eq. "t") .OR. &
            (trim(adjustl(cmasktype)) .eq. "smsl") .OR. &
            (trim(adjustl(cmasktype)) .eq. "hmsl") .OR. &
            (trim(adjustl(cmasktype)) .eq. "tmsl")) then
          write(errstr,*)" you chose your cheesemasktype wisely..."
          call message(AppMsg,Noisy)
       else
          write(errstr,*)"cheesemasktype can only be s, h, t, smsl, hmsl, or tmsl"
          call error("badCheeseMaskType",trim(adjustl(errstr)))
       endif
    endif
    cmos = .FALSE.
    if (type .eq. 6) then ! cheese mosaic
       cmostype = stringParameter(cmospar)
       cmos = .TRUE.
       write(errstr,*)'Mosaic cheese mask type:         ',cmostype
       call message(AppMsg,Always)
       if ((trim(adjustl(cmostype)) .eq. "s") .OR. (trim(adjustl(cmostype)) .eq. "h") .OR. &
            (trim(adjustl(cmostype)) .eq. "t") ) then
          write(errstr,*)" you chose your cheesetype wisely..."
          call message(AppMsg,Noisy)
       else
          write(errstr,*)"type=6 cheesemosaictype can only be s, h, t"
          call error("badCheeseMosaicType",trim(adjustl(errstr)))
       endif
       if (withchee) then
          call warning("cheeseCheese",&
               "You are attempting to apply a cheese mask to mosaicked cheese masks.")
       endif
    endif
    if (type .eq. 7) then ! cheese merged source list (msl) mosaic
       cmostype = stringParameter(cmospar)
       cmos = .TRUE.
       write(errstr,*)'Mosaic cheese mask type:         ',cmostype
       call message(AppMsg,Always)
       if ((trim(adjustl(cmostype)) .eq. "smsl") .OR. (trim(adjustl(cmostype)) .eq. "hmsl") &
            .OR. (trim(adjustl(cmostype)) .eq. "tmsl")) then
          write(errstr,*)" you chose your cheesetype wisely..."
          call message(AppMsg,Noisy)
       else
          write(errstr,*)"type=7 cheesemosaictype can only be msls, mslh, or mslt"
          call error("badCheeseMosaicType",trim(adjustl(errstr)))
       endif
       if (withchee) then
          call warning("cheeseCheese",&
               "You are attempting to apply a cheese mask to mosaicked cheese masks")
       endif
    endif
    ! cast ints to char
    write(celow,'(I5)')  elow
    write(cehigh,'(I5)') ehigh
    write(cxdim,'(I5)')  xdim
    write(cydim,'(I5)')  ydim
    !
    call makeOutName(types(type),cmos,cmostype,celow,cehigh,out_F) ! e.g. mosaic-protimsky-350-1100.fits
    die = .TRUE.
    call check_clobber(out_F,die)
    !
    ! BHP: write out all the input params and derived values
    call printParams(out_F,alpha,coordsys,ctypeN,crvalN,celow,cehigh,type,types,cxdim,cydim,pixSize)
    !
    die = .TRUE.
    call check_clobber(trim(adjustl(out_F)),die)
    !
    allocate(mosaic(xdim,ydim))
    mosaic = 0.
    !
    ! SS: Get unused luns. BHP: taken care of with check_open_set below
    !
    djul = 2451544.5D0
    if(coordsys(1:2) .eq. 'ec') then ! type=ecliptic=1
       dlin = drad*crvaln(1)
       dbin = drad*crvaln(2)
       call equecl(djul,.FALSE.,.FALSE.,dlin,dbin,dlout,dbout)
       dracen  = dlout/drad
       ddeccen = dbout/drad
       !
       ! BHP: old ierr handled in goflib/equecl
       ! SS:  Leave in equatorial coordinates
       !
    elseif(coordsys(1:2) .eq. 'eq') then ! type=equitorial=2
       dracen  = crvaln(1)
       ddeccen = crvaln(2)

       ! SS: Transform from galactic to equatorial coordinates

    elseif(coordsys .eq. 'ga') then ! type=galactic=3
       dlin = drad*crvaln(1)
       dbin = drad*crvaln(2)
       call equgal(djul,.FALSE.,dlin,dbin,dlout,dbout)
       dracen  = dlout/drad
       ddeccen = dbout/drad

    endif
    dracen = drad*dracen
    ddeccen = drad*ddeccen
    write(errstr,*)' dracen   = ',dracen
    call message(AppMsg,Noisy)
    write(errstr,*)' ddeccen  = ',ddeccen
    call message(AppMsg,Noisy)

    ! SS: Create output file name. BHP: done above in call makeoutfile

    ! SS: Set the maximum separation for a pointing to still possibly be in the image

    field = 0.0174533d0*(pixSize*sqrt(1.0*(xdim*xdim+ydim*ydim))/2.0 + 15.0)/60.0

    ! BHP: 'linn' variable declared but never USED, so eliminated

    crpixN(1) = xdim/2.0 + 0.5
    crpixN(2) = ydim/2.0 + 0.5
    cdeltN(1) = -pixSize/60.0 ! OLDNEW: scale/pixSize
    cdeltN(2) = pixSize/60.0
    crotn = 0.0 ! why???!!!
    !
    ! SS: Set the band
    !
    islow = (elow + 25)/50
    ishigh = (ehigh + 25)/50
    !
    ! SS: Set the input file name. BHP: check, open, read dir.list ASCII file
    lun = GetLun()
    inquire(FILE=trim(adjustl(dir_F)),EXIST=exists)
    if (.not.exists) call error("noDirfile","Input dirfile does not exist")
    open(unit=lun,status='old',form='formatted',file=trim(adjustl(dir_F)),iostat=ios)
    ios       = 0
    nPrefixes = 0
    nRows     = 0
    allocate(prefixes(5000),dirs(5000)) ! BHP: change this???!!!
1020 format(A1024)
    !
    ! BHP: loop through all dir.list files, separate prefixes from dir paths
    !
    OUTER: do
       lastnn = 0
       read(lun,1020,iostat=ios) curdir ! current directory
       if (ios .ne. 0) exit
       nRows = nRows + 1
       nn = index(trim(adjustl(curdir)),"/")
       ctemp = curdir(nn+1:)
       INNER: do
          nn = index(trim(adjustl(ctemp)),"/")
          lastnn = lastnn + nn
          if (nn .eq. index(curdir," ")) then
             exit INNER
          else
             ctemp = ctemp(nn+1:)
             if (nn .eq. 0) then
                dirs(nRows) = curdir(:lastnn)
                prefixes(nRows) = trim(adjustl(ctemp))
                write(errstr,*)nRows," curdir: ",curdir(:lastnn),' prefix = ',trim(adjustl(ctemp)), &
                     lastnn,' len = ',index(curdir," ")-1
                call message(AppMsg,Noisy)
                nPrefixes = nPrefixes + 1
                exit INNER
             endif
          endif
       end do INNER
    end do OUTER
    if (nPrefixes .eq. 0) then
       call error("noPrefixes","None of your dirfile prefixes exist, exiting...")
    endif
    write(errstr,*)'Initial (input) nPrefixes = ',nPrefixes,' nRows = ',nRows
    call message(AppMsg,Always)
    !
    ! BHP: loop through input files and accumulate the mosaic
    !
    allocate(mask(xGood,yGood)) ! create and initialize mask for below
    allocate(fovim(xGood,yGood),fovimoot(xGood,yGood))
    fovim     = 0.
    fovimoot  = 0.
    mask      = 1.
    nPrefixes = 0
    !
    do i=1,nRows
       write(errstr,*)"ADD NEW ARRAY"
       call message(AppMsg,Always) ! change???!!!
       pn = .FALSE.
       exists = .FALSE.
       call makeInName(prefixes(i),types(type),cmos,cmostype,celow,cehigh,basename,baseoot)
       in_F    = trim(adjustl(dirs(i))) // "/" // trim(adjustl(basename))
       inOOT_F = trim(adjustl(dirs(i))) // "/" // trim(adjustl(baseoot))
       write(errstr,*)'Current input filename = ',trim(adjustl(in_F)) ! full path to current infile
       call message(AppMsg,Always) ! change to Noisy???!!!
       inquire(FILE=trim(adjustl(in_F)),EXIST=exists)
       if (.not.exists) then
          write(errstr,*)"Current input file : ",trim(adjustl(in_F))," does not exist, skipping"
          call warning("noCurfile",errstr)
          cycle
       endif
       ! SS: Open the FITS file for the component (BHP: OLDNEW: type)
       status = 0
       warndie = .FALSE.
       call check_open_set(in_F,in_S,1,status,warndie)
       if (status .eq. 0) then
          write(errstr,*)"Current input file : ",trim(adjustl(in_F))," cannot be opened or read, skipping"
          call warning("badInfile",errstr)
          cycle
       endif
       nPrefixes = nPrefixes + 1
       !
       ! SS: Read the FITS file header information
       !
       status = 0
       warndie = .FALSE. ! mosaicmerge won't die if can't read singular inputs
       typeF = 3         ! 1 = EVLI, 2 = SPEC, 3 = IMG
       call read_header_info(typeF,in_S,out_T,out_A,instO,obsidO,expidO,smodeO,&
            frmtimeO,filtO,revlO,backO,areaO,expoO,elowO,ehighO,nRowsO,warndie)
       inst = 'EMOS1' ! pseudo 'default'
       if (index(instO,"M2") .ne. 0) then
          inst = 'EMOS2'
       elseif (index(instO,"PN") .ne. 0) then
          if (type .eq. 1) pn = .TRUE.
          inst = 'EPN'
       endif
       warndie = .FALSE.
       ! SS: Get the useful keywords
       ! SS: And a few more
       call coord_keys (in_S,ra_pntO,dec_pntO,pa_pntO,crvalO,cdeltO,cunitO,ctypeO,&
            crpixO,crotaO,radecsysO,warndie)
       dpra  = drad*crvalO(1)
       dpdec = drad*crvalO(2)
       !
       ! SS: Test to see how far the pointing is from the center of the mosaic
       !
       delang = 0.
       call angsep(dracen,ddeccen,dpra,dpdec,delang)
       write(errstr,*)'ddeccen = ',ddeccen
       call message(AppMsg,Verbose) ! change???!!!
       write(errstr,*)'dpra    = ',dpra
       call message(AppMsg,Verbose) ! change???!!!
       write(errstr,*)'dpdec   = ',dpdec
       call message(AppMsg,Verbose) ! change???!!!
       write(errstr,*)'dlang   = ',delang
       call message(AppMsg,Verbose) ! change???!!!
       !
       if(delang .lt. field) then
          !
          ! SS:  Set the scale factor for the filter
          fscal = 1.0
          ! OLD: call hard_scale(xmmdir,alpha,idet,filt,islow,ishigh,fscal)
          if (type .eq. 2) then 
             call hardscale(inst,filtO,alpha,islow,ishigh,fscal) ! goflib/esas_mod
             itot = itot + 1
             if(index(filtO,'Thin') .ne. 0) then
                ithin = ithin + 1
             elseif(index(filtO,'Medi') .ne. 0) then
                imed = imed + 1
             elseif(index(filtO,'Thic') .ne. 0) then
                ithick = ithick + 1
             endif
          endif
          write(errstr,*)'Filter scale factor = ',fscal
          call message(AppMsg,Always) ! change???!!!
          !
          ! SS: Print the central coords of the observation
          ! BHP: see here:
          ! https://heasarc.gsfc.nasa.gov/docs/software/fitsio/c/f_user/node58.html
          ! it's ALWAYS 8 characters and ftwldp expects C*4 to be one of:
          ! -SIN, -TAN, -ARC, -NCP, -GLS, -MER, and -AIT
          ! cf. https://danmoser.github.io/notes/gai_fits-imgs.html
          !
          ctyp = ctypeO(1)(5:8)
          write(errstr,*)'Original image coordinate system: ',ctyp
          call message(AppMsg,Always) ! change???!!!
          write(errstr,*)'Original image coordinate ID: ', &
               ctypeO(1), '  ', ctypeO(2)
          call message(AppMsg,Always) ! change???!!!
          write(errstr,*)'Original image central coordinates: ', &
               crvalO(1), crvalO(2)
          call message(AppMsg,Always) ! change???!!!
          write(errstr,*)'Original image reference pixel: ', &
               crpixO(1), crpixO(2)
          call message(AppMsg,Always) ! change???!!!
          write(errstr,*)'Original image pixel sizes: ', &
               cdeltO(1), cdeltO(2)
          call message(AppMsg,Always) ! change???!!!
          write(errstr,*)'Original image rotation angle: ',crotaO
          call message(AppMsg,Always) ! change???!!!
          !
          ! SS:  Read in the X-ray image
          ! BHP: fovim and cheese are Int32, all others are Real32, cast to Real32
          ! check msl???!!!
          !
          mask  = 1. ! level it out each time
          fovim = 0. ! zero out current image
          if (withmask) then
             !
             ! BHP: get the current directory, current prefix, ordinary mask equivalent
             !      e.g. mos1S001-maskimsky-350-1100.fits
             call makeInName(prefixes(i),'mask',.FALSE.,cmostype,celow,cehigh,basename,baseoot)
             mask_F = trim(adjustl(dirs(i))) // "/" // trim(adjustl(basename))
             call getMask(mask_F,xGood,yGood,cmos,mask)
          endif
          if (withchee) then
             !
             ! BHP: get the current directory, current prefix, cheese mask equivalent
             !      e.g. mos1S001-cheeset.fits
             call makeInName(prefixes(i),'cheese',.TRUE.,cmasktype,celow,cehigh,basename,baseoot)
             mask_F = trim(adjustl(dirs(i))) // "/" // trim(adjustl(basename))
             call getMask(mask_F,xGood,yGood,.TRUE.,mask)
          endif
          if ((type .eq. 1) .or. (type .eq. 6) .or. (type .eq. 7))  then 
             call read_2Darray(in_S,in_A,aDR32,aDR64,aDI16,fovim_JP,xSize,ySize,typeA)
             do j=1,xSize
                do jj=1,ySize
                   if((fovim_JP(j-1,jj-1) .ne. 0.0) .and. (mask(j,jj) .gt. 0)) then
                      fovim(j,jj) = real(fovim_JP(j-1,jj-1))
                   endif
                enddo
             enddo
          else
             call read_2Darray(in_S,in_A,fovim_RP,aDR64,aDI16,aDI32,xSize,ySize,typeA)
             do j=1,xSize
                do jj=1,ySize
                   if((fovim_RP(j-1,jj-1) .ne. 0.0) .and. (mask(j,jj) .gt. 0)) then
                      fovim(j,jj) = fovim_RP(j-1,jj-1)
                   endif
                enddo
             enddo
          endif
999       format(i4,1x,i4,1x,e15.6)
          write(errstr,*)'Original image NAXES, NAXIS: ',xSize,ySize
          call message(AppMsg,Always) ! change???!!!
          !
          if ((xSize /= xGood) .or. (ySize /= yGood)) then
             write(errstr,*)"OBJ image dims not: ",xGood,"x",yGood
             call error("wrongArrayDims",trim(errstr))
          end if
          ! set array in Real32 for use later
          ! fovim = 0.
          ! SS:  Close the FITS file
          call release(in_S)
          ! SS:  Read in the OOT event file if pn data
          ! BHP: should probably sanity check OOT vs PN coordkeys, headerkeys???!!!
          if (pn) then
             if (.not.cmos) then
                write(errstr,*)'Current OOT input filename = ',trim(adjustl(inOOT_F)) ! full path to current infile
                call message(AppMsg,Always) ! change to Noisy???!!!
                inquire(FILE=trim(adjustl(inOOT_F)),EXIST=exists)
                if (.not.exists) then ! what to do about this???!!!
                   write(errstr,*)"Current input OOT file, ",trim(adjustl(inOOT_F))," does not exist, skipping."
                   call warning("noCurfile",errstr)
                   cycle
                endif
                status = 0
                warndie = .FALSE.
                call check_open_set(inOOT_F,inOOT_S,1,status,warndie)
                if (status .eq. 0) then
                   write(errstr,*)"Current input OOT file, ",trim(adjustl(inOOT_F))," cannot be opened or read, skipping"
                   call warning("noOOTInfile",errstr)
                   cycle
                endif
                status = 0
                warndie = .FALSE. ! mosaicmerge won't die if can't read singular inputs
                typeF = 3         ! 1 = EVLI, 2 = SPEC, 3 = IMG
                call read_header_info(typeF,inOOT_S,inOOT_T,inOOT_A,instOO,obsidOO,expidOO,smodeOO,&
                     frmtimeOO,filtOO,revlOO,backOO,areaOO,expoOO,elowOO,ehighOO,nRowsOO,warndie)
                warndie = .FALSE.
                call getootsca(smodeO,ootsca,warndie) ! goflib/gofutils
                ! SS:  Read in the X-ray image
                call read_2Darray(inOOT_S,inOOT_A,aDR32,aDR64,aDI16,fovimoot_JP,&
                     xSize,ySize,typeA)
                if ((xSize /= xGood) .or. (ySize /= yGood)) then
                   write(errstr,*)"OBJ OOT image dims not: ",xGood,"x",yGood
                   call error("wrongArrayDims",trim(errstr))
                end if
                ! set array in Real32 for use later
                !fovimoot = 0.
                do j=0,xSize-1 ! pointers from 0, arrays from 1
                   do jj=0,ySize-1
                      fovimoot(j+1,jj+1) = real(fovimoot_JP(j,jj))
                      !write(51,999)j,jj,fovimoot(j+1,jj+1)
                   enddo
                end do
                ! SS: Close the FITS file
                call release(inOOT_S)
                ! SS: Combine the data
                do ii=1,xGood
                   do jj=1,yGood
                      fovim(ii,jj) = fovim(ii,jj) - ootsca*fovimoot(ii,jj)
                      !write(52,999)j,jj,fovim(j+1,jj+1)
                   enddo
                enddo
             endif
          endif
       endif
       !
       ! SS: Cast the new observation into the mosaic
       !
       cast = .TRUE.
       do ii=1,xGood ! OLDNEW: 900/xGood
          xpixin = ii
          do jj=1,yGood ! OLDNEW: 900/yGood
             ypixin = jj
             cra    = 0.
             cdec   = 0.
             status = 0
             if((fovim(ii,jj) .ne. 0.0) .and. (mask(ii,jj) .gt. 0)) then
                ! ftwldp is in fitsio
                call ftwldp(xpixin,ypixin,crvalO(1),crvalO(2),&
                     crpixO(1),crpixO(2),cdeltO(1),cdeltO(2), &
                     crotaO(1),ctyp,cra,cdec,status)
                if(status .ne. 0) then
                   write(errstr,*)"Bad return from ftwldp, pixel not cast"
                   call message(AppMsg,Noisy) ! change to Noisy???!!!
                   write(errstr,*)"Occurred in file: ",trim(adjustl(in_F))
                   call message(AppMsg,Noisy) ! change to Noisy???!!!
                   cast = .FALSE.
                endif
                !
                ! SS: Transform from equatorial to ecliptic coordinates
                !
                if(coordsys .eq. 'ec') then ! Transform from equatorial to ecliptic coordinates
                   dlin = drad*cra
                   dbin = drad*cdec
                   call equecl(djul,.FALSE.,.TRUE.,dlin,dbin,dlout,dbout)
                   cra  = dlout/drad
                   cdec = dbout/drad
                elseif(coordsys .eq. 'eq') then ! SS: Leave in equatorial coordinates
                   continue
                elseif(coordsys .eq. 'ga') then ! SS: Transform from equatorial to galactic coordinates
                   dlin = drad*cra
                   dbin = drad*cdec
                   call equgal(djul,.TRUE.,dlin,dbin,dlout,dbout)
                   cra  = dlout/drad
                   cdec = dbout/drad
                else
                   call error("badCoordsys","Coordsys must be 1/ec, 2/eq, or 3/ga") ! should never happen!
                endif
                !
                ! SS:  Transform into the merged system
                ! ftxypx is in fitsio
                call ftxypx (cra,cdec,crvalN(1),crvalN(2),crpixN(1),crpixN(2), &
                     cdeltN(1),cdeltN(2),crotN,ctyp,pixcrd(1),pixcrd(2),status)
                ipixcrd(1) = nint(pixcrd(1))
                ipixcrd(2) = nint(pixcrd(2))
                cast = .TRUE.
                if(status .ne. 0) then
                   write(errstr,*)"Bad return from ftxypx, pixel not cast"
                   call message(AppMsg,Noisy) 
                   write(errstr,*)"Occurred in file: ",trim(adjustl(in_F))
                   call message(AppMsg,Noisy) 
                   cast = .FALSE.
                endif
                if (cast) then
                   if((ipixcrd(1) .ge. 1) .and. &
                        (ipixcrd(1) .le. xdim) .and. &
                        (ipixcrd(2) .ge. 1) .and. &
                        (ipixcrd(2) .le. ydim)) then
                      mosaic(ipixcrd(1),ipixcrd(2)) = &
                           mosaic(ipixcrd(1),ipixcrd(2)) + &
                           fscal*fovim(ii,jj)
                   endif
                endif
             endif
          enddo
       enddo
    enddo
    if (nPrefixes .eq. 0) then
       call error("noPrefixes","None of your dirfile prefixes exist, exiting...")
    endif
    ! writeMap is in goflib/gofio_mod
    ! call writeMap(temp_F,out_F,fovim,xDim,yDim)
    ! SS: Write out the merged array
    !
    !    type: Component Selection
    !       1 => count image
    !       2 => exposure
    !       3 => QPB counts
    !       4 => SP counts
    !       5 => SWCX counts
    !       6 => Cheese masks from cheese or cheese-bands
    !       7 => Cheese masks from make_mask_merge
    !
    units   = '        '
    comment = '        '
    if(type .eq. 1) then
       units = 'counts'
       comment = 'Merged observed count image'
    elseif(type .eq. 2) then
       units = 's'
       comment = 'Merged exposure map, for true scale by in/out pixel sizes'
    elseif(type .eq. 3) then
       units = 'counts'
       comment = 'Merged model QPB counts'
    elseif(type .eq. 4) then
       units = 'counts'
       comment = 'Merged model SP counts'
    elseif(type .eq. 5) then
       units = 'counts'
       comment = 'Merged model SWCX counts'
    elseif(type .eq. 6) then
       units = 'None'
       comment = 'Mask values 1=include, 0=not'
    elseif(type .eq. 7) then
       units = 'None'
       comment = 'Mask values 1=include, 0=not'
    endif
    bzero  = 0.0
    bscale = 1.0
    status = 0
    write(errstr,*) 'Merged image coordinate system: ', ctyp
    call message(AppMsg,Always)
    write(errstr,*) 'Merged image coordinate ID: ', &
         ctypen(1), '  ', ctypen(2)
    call message(AppMsg,Always)
    write(errstr,*) 'Merged image central coordinates: ', &
         crvalN(1), crvalN(2)
    call message(AppMsg,Always)
    write(errstr,*) 'Merged image reference pixel: ', &
         crpixN(1), crpixN(2)
    call message(AppMsg,Always)
    write(errstr,*) 'Merged image pixel sizes: ', &
         cdeltN(1), cdeltN(2)
    call message(AppMsg,Always)
    write(errstr,*) 'Merged image rotation angle: ', CROTN
    call message(AppMsg,Always)
    write(errstr,*) 'Cast type = ', types(type)
    call message(AppMsg,Always)
    ! putfit is in goflib/gofio, BHP fix clobber ???!!!
    call putfit(out_f,mosaic,ini4l,1,units,comment,xdim,ydim,&
         bzero,bscale,ctypeN,crvalN,cdeltN,crpixN,clobber,status)
    !
    if (type .eq. 2) then
       write(errstr,888) 'Total Observations         : ', ithin+imed+ithick
       call message(AppMsg,Always)
       write(errstr,888) 'Thin Filter Observations   : ', ithin
       call message(AppMsg,Always)
       write(errstr,888) 'Medium Filter Observations : ', imed
       call message(AppMsg,Always)
       write(errstr,888) 'Thick Filter Observations  : ', ithick
       call message(AppMsg,Always)
    endif

888 format(a,1x,i5)

    return

  end subroutine mosaicmerge

  subroutine printParams(out_F,alpha,coordsys,ctype,crval,celow,cehigh,type,types,cxdim,cydim,pixelsize)
    !
    ! SS: List the program control parameters.
    !
    character(len=*) :: out_F, ctype(2), celow, cehigh, coordsys, cxdim, cydim, types(7)
    integer :: coord, type, xdim, ydim
    real(kind=single) :: alpha, pixelsize
    real(kind=double) :: crval(2)
    !
    write(errstr,*)"Output filename: ",trim(adjustl(out_F))
    call message(AppMsg,Always)
    if((coordsys .eq. '1') .or. (coordsys(1:2) .eq. 'ec'))  then
       ctype(1) = 'ELON-TAN'
       ctype(2) = 'ELAT-TAN'
    elseif((coordsys .eq. '2') .or. (coordsys(1:2) .eq. 'eq')) then
       ctype(1) = 'RA---TAN'
       ctype(2) = 'DEC--TAN'
    elseif((coordsys .eq. '3') .or. (coordsys(1:2) .eq. 'ga')) then
       ctype(1) = 'GLON-TAN'
       ctype(2) = 'GLAT-TAN'
    endif
    write(errstr,*)"Assumed Spectral Index:       : ",alpha
    call message(AppMsg,Always)
    write(errstr,*)"Coordinate System:            : ",ctype(1)," ",ctype(2)
    call message(AppMsg,Always)
    write(errstr,*)"Central Coordinates           : ",crval(1)," ",crval(2)
    call message(AppMsg,Always)
    write(errstr,*)"Energy Range (elow/ehigh)     : ",trim(adjustl(celow)),"-",trim(adjustl(cehigh))
    call message(AppMsg,Always)
    write(errstr,*)"Output array size (xdim/ydim) : ",trim(adjustl(cxdim))," x ",trim(adjustl(cydim))
    call message(AppMsg,Always)
    write(errstr,*)"Pixel size                    : ",pixelsize
    call message(AppMsg,Always)
    write(errstr,*)"Cast type                     : ",types(type)
    call message(AppMsg,Always)

    return

  end subroutine printParams

  subroutine makeInName(prefix,type,cmos,ctype,celow,cehigh,in_F,inOOT_F)
    character(len=*) :: prefix
    character(len=256), intent(out) :: in_F, inOOT_F
    character(len=*) :: ctype, type, celow, cehigh
    logical :: cmos
    !
    ! makes out_F, e.g.s.:
    ! prefix-fovim-350-1100.fits
    ! prefix-bkgim-350-1100.fits
    ! prefix-expim-350-1100.fits
    ! prefix-protim-350-1100.fits
    ! prefix-swcxim-350-1100.fits
    ! prefix-cheese(s/h/t)-350-1100.fits
    ! prefix-cheese(s/h/t)msl-350-1100.fits
    !
    if (cmos) then ! mosaicking cheese masks
       ! e.g., mos1S001-cheeseh.fits, pnS003-cheesetmsl.fits, etc
       in_F = trim(adjustl(prefix)) // "-cheese" // trim(adjustl(ctype)) // ".fits"
       inOOT_F = "" ! not used but make sure it's initialized...
    else
       in_F = trim(adjustl(prefix)) // "-" // trim(adjustl(type)) // "imsky-" // &
            trim(adjustl(celow)) // "-" // trim(adjustl(cehigh)) // ".fits"
       inOOT_F = trim(adjustl(prefix)) // "-" // trim(adjustl(type)) // "imootsky-" // &
            trim(adjustl(celow)) // "-" // trim(adjustl(cehigh)) // ".fits"
    endif

    return

  end subroutine makeInName
  !
  !
  !
  subroutine makeOutName(type,cmos,ctype,celow,cehigh,out_F)
    character(len=256), intent(out) :: out_F
    character(len=*) :: ctype, type, celow, cehigh
    logical :: cmos
    !
    ! makes out_F, e.g.s.:
    ! mosaic-fovim-350-1100.fits
    ! mosaic-bkgim-350-1100.fits
    ! mosaic-expim-350-1100.fits
    ! mosaic-protim-350-1100.fits
    ! mosaic-swcxim-350-1100.fits
    ! mosaic-cheeset.fits
    ! mosaic-cheesehmsl.fits
    ! etc
    if (cmos) then
       out_F =  "mosaic-cheese" // trim(adjustl(ctype)) // ".fits"
    else
       out_F = "mosaic-" // trim(adjustl(type)) // "im-" // &
            trim(adjustl(celow)) // "-" // trim(adjustl(cehigh)) // ".fits"
    endif

    return

  end subroutine makeOutName

  subroutine getMask(in_F,xGood,yGood,cmos,mask)

    type(DataSetT) :: in_S
    type(ArrayT)   :: in_A
    character(len=*) :: in_F
    integer, dimension(:), pointer :: dimVector
    integer :: status, i, j, xGood, yGood, xSize, ySize, typeA
    integer(kind=int32), dimension(:,:), pointer :: temp_JP
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    logical :: cmos, warndie
    real(kind=single), dimension(:,:) :: mask
    real(kind=single), dimension(:,:), pointer :: aDR32
    real(kind=double), dimension(:,:), pointer :: aDR64
    !
    ! SS: Read in point source masks to go with the observation
    !
    mask    = 1.
    status  = 0
    warndie = .FALSE.
    call check_open_set(trim(adjustl(in_F)),in_S,1,status,warndie) ! won't die if there's a problem
    ! masks are in the MASK extension                                        
    if (status .eq. 0) then
       call warning("badMask","Input MASK file cannot be opened or read, skipping")
       return
    endif
    ! 
    if (cmos) then ! if cheese masking, get the array from the PRIMARY
       call read_2Darray(in_S,in_A,aDR32,aDR64,aDI16,temp_JP,xSize,ySize,typeA)
    else
       ! if ordinary masking, get the array from the cheese MASK extension
       in_A = array(in_S,"MASK",READ) ! from eexpmap                            
       dimVector => dimensions(in_A)
       xSize = dimVector(0)
       ySize = dimVector(0)
       temp_JP   => int32Array2Data(in_A)
    endif
    ! sanity check
    if ((xSize .ne. xGood) .or. (ySize .ne. yGood)) then
       write(errstr,666)'Input array must be ',xGood,'x',yGood
       call error('wrongArrayDims',trim(adjustl(errstr)))
    end if
    !
666 format(A,i4,A,i4)
    do i=1,xSize     ! 900                                                      
       do j=1,ySize  ! 900                                                      
          mask(i,j) = temp_JP(i-1,j-1)
       enddo
    enddo
    nullify(temp_JP)
    call release(in_S)

    return
  end subroutine getMask

end module mosaicmerge_mod
